#!/bin/bash
############################################################################
# (C) Copyright 2016-2017 Hewlett Packard Enterprise Development LP.
# @(#) Serviceguard Easy Deployment for Microsoft SQL Server on Linux.
# @(#) Product Name                :  HPE Serviceguard extension for Microsoft SQL Server. SGeMSSQL
# @(#) Product Version             :  A.12.20.00
# @(#) Patch Name                  :  
#%%DO_NOT_MODIFY_TEXT%%
#
############################################################################

#
# Get parameters
#
if [[ $# == 0 ]]
then
   sg_log 0 "ERROR: No parameter is passed to the command."
   sg_exit 1 $*
fi
. /etc/cmcluster.conf

SG_PACKAGE=$1

#
# Source in the utility functions used by the module scripts
#
if [[ -z $SG_UTILS ]]
then
    SG_UTILS=$SGCONF/scripts/mscripts/utils.sh
fi
if [[ -f ${SG_UTILS} ]]
then
    . ${SG_UTILS}
    if [[ $? -ne 0 ]]
    then
        echo "ERROR: Unable to source package utility functions file: ${SG_UTILS}"
        exit 1
    fi
else
    echo "ERROR: Unable to find package utility functions file: ${SG_UTILS}"
    exit 1
fi

SGEMSSQL_UTILS=${SGCONF}/scripts/sgemssql/mssql_utils.sh
export SGEMSSQL_UTILS
if [[ -f ${SGEMSSQL_UTILS} ]]; then
    . ${SGEMSSQL_UTILS}
    if (( $? != 0 ))
    then
        echo "ERROR: Unable to source SGeMSSQL utility functions file: ${SGEMSSQL_UTILS}"
        exit 1
    fi
else
    echo "ERROR: Unable to find SGeMSSQL utility functions file: ${SGEMSSQL_UTILS}"
    exit 1
fi

unset SG_ENV_FILE
sg_source_pkg_env $1

#
# PATH for Linux
#
PATH=$SGSBIN:/usr/bin:/usr/sbin:/etc:/bin:/sbin:/usr/lbin
export PATH

#$AG $max_seq_number $max_node_id $my_role
AG=$2
caller_max_seq_num=$3
caller_max_node_id=$4
caller_role=$5

sg_log 0 "INFO: Received request to validate if remote node"\
         "\"`get_node_name_from_nodeid $caller_max_node_id`\" is eligible to become a PRIMARY replica"

if [ $caller_role -eq $PRIMARY ]; then
    sg_log 0 "INFO: Response from node \"$LOCAL_HOST_NAME\" to remote node"\
             "\"`get_node_name_from_nodeid $caller_max_node_id`\" is that, node"\
             "\"`get_node_name_from_nodeid $caller_max_node_id`\" is eligible as it"\
             "is already a PRIMARY REPLICA."
    echo "$TRUE"
    exit 0
fi

# First get admin user and password
get_user_name_and_password_from_login_credentials_file
if [ $? -ne 0 ]; then
    sg_log 0 "ERROR: Response from node \"$LOCAL_HOST_NAME\" to remote node"\
             "\"`get_node_name_from_nodeid $caller_max_node_id`\" is that the"\
             "MSSQL Credential validation failed on the node \"$LOCAL_HOST_NAME\"."
    return "$TRUE"
    exit 0
fi

fname="sgemssql_foot_print_file_"$AG".txt"
foot_print_file="$SGRUN/sgemssql/$fname"

get_primary_replica_identification_params $AG
get_the_role_of_the_ag_on_local_replica $AG
local_ag_role=$?

if [ $local_ag_role -eq $PRIMARY ]; then
    sg_log 0 "INFO: Response from node \"$LOCAL_HOST_NAME\" to remote node"\
             "\"`get_node_name_from_nodeid $caller_max_node_id`\" is that, node"\
             "\"`get_node_name_from_nodeid $caller_max_node_id`\" is not eligible as"\
             "node \"$LOCAL_HOST_NAME\" is already a PRIMARY REPLICA."
    echo "$FALSE"
    exit 0
fi

are_all_db_in_healthy_state_for_a_node $AG $LOCAL_HOST_NAME
db_health=$?
if [[ $db_health -eq $FALSE ]]; then
    sg_log 0 "INFO: Response from node \"$LOCAL_HOST_NAME\" to remote node"\
             "\"`get_node_name_from_nodeid $caller_max_node_id`\" is that, node"\
             "\"`get_node_name_from_nodeid $caller_max_node_id`\" is eligible as one of the"\
             "Database on node \"$LOCAL_HOST_NAME\" is unhealthy."
    echo "$TRUE"
    exit 0
fi

# Check the state of db synchronization state from the file. If the db
# synchronization_health is not healthy, this node is not eligible.
if [[ -s $foot_print_file ]]; then
    get_synchronization_health_from_file $foot_print_file $LOCAL_HOST_NAME
    _is_db_sync_state_healthy=$?
    if [[ $_is_db_sync_state_healthy -eq $FALSE ]]; then
        sg_log 0 "INFO: Response from node \"$LOCAL_HOST_NAME\" to remote node"\
             "\"`get_node_name_from_nodeid $caller_max_node_id`\" is that, node"\
             "\"`get_node_name_from_nodeid $caller_max_node_id`\" is eligible as one(or more) of"\
             "the Database(s) synchronization_health on node \"$LOCAL_HOST_NAME\" is not healthy."
        echo "$TRUE"
        exit 0
    fi
fi


if [[ $seq_num_from_db -lt $caller_max_seq_num ]]; then
    # The caller is biggest and greatest
    sg_log 0 "INFO: Response from node \"$LOCAL_HOST_NAME\" to remote node"\
             "\"`get_node_name_from_nodeid $caller_max_node_id`\" is that, the node"\
             "\"`get_node_name_from_nodeid $caller_max_node_id`\" is eligible as the"\
             "sequence_number on node \"$LOCAL_HOST_NAME\" is lesser than that on node"\
             "\"`get_node_name_from_nodeid $caller_max_node_id`\"."
    echo "$TRUE"
    exit 0
elif [[ $seq_num_from_db -gt $caller_max_seq_num ]]; then
    # The caller is not uptodate
    sg_log 0 "INFO: Response from node \"$LOCAL_HOST_NAME\" to remote node"\
             "\"`get_node_name_from_nodeid $caller_max_node_id`\" is that, the node"\
             "\"`get_node_name_from_nodeid $caller_max_node_id`\" is not eligible as the"\
             "sequence_number on node \"$LOCAL_HOST_NAME\" is greater than that on node"\
             "\"`get_node_name_from_nodeid $caller_max_node_id`\"."
    echo "$FALSE"
    exit 0
elif [[ $seq_num_from_db -eq $caller_max_seq_num ]]; then

    # 1. Check if there is any lag in seconds for this node
    sg_log 0 "INFO: Both the nodes have same sequence number."
    do_i_lag_in_sec_from_file $foot_print_file $LOCAL_HOST_NAME
    do_i_lag=$?
    if [ $do_i_lag -eq $TRUE ]; then
        sg_log 0 "INFO: Response from node \"$LOCAL_HOST_NAME\" to remote node"\
             "\"`get_node_name_from_nodeid $caller_max_node_id`\" is that, the node"\
             "\"`get_node_name_from_nodeid $caller_max_node_id`\" is eligible as one of the"\
             "Database on node \"$LOCAL_HOST_NAME\" lags."
        echo "$TRUE"
        exit 0
    fi

    # 2. Check if local alone is RESOLVING, caller will give up
    if [ $local_ag_role -eq $RESOLVING ] && [ $caller_role -eq $SECONDARY ]; then
        sg_log 0 "INFO: Response from node \"$LOCAL_HOST_NAME\" to remote node"\
                 "\"`get_node_name_from_nodeid $caller_max_node_id`\" is that, the node"\
                 "\"`get_node_name_from_nodeid $caller_max_node_id`\" is not eligible"\
                 "as the node \"$LOCAL_HOST_NAME\" is RESOLVING and node"\
                 "\"`get_node_name_from_nodeid $caller_max_node_id`\" is not RESOLVING."
        echo "$FALSE"
        exit 0
    fi

    # 3. Check if Caller alone is RESOLVING, local will give up
    if [ $local_ag_role -eq $SECONDARY ] && [ $caller_role -eq $RESOLVING ]; then
        sg_log 0 "INFO: Response from node \"$LOCAL_HOST_NAME\" to remote node"\
                 "\"`get_node_name_from_nodeid $caller_max_node_id`\" is that, the node"\
                 "\"`get_node_name_from_nodeid $caller_max_node_id`\" is eligible as the"\
                 "node \"`get_node_name_from_nodeid $caller_max_node_id`\" is RESOLVING and node"\
                 "\"$LOCAL_HOST_NAME\" is not RESOLVING."
        echo "$TRUE"
        exit 0
    fi

    # 4. If both are RESOLVING and both are having same seq_num, the lowest node-id will win the race
    if [ $local_ag_role -eq $RESOLVING ] && [ $caller_role -eq $RESOLVING ]; then
        # 2.1 Evaluate the lowest nodeid
        sg_log 0 "INFO: Both the nodes are RESOLVING."
        get_nodeid_from_nodename $LOCAL_HOST_NAME
        local_node_id=$?
        if [ $local_node_id -lt $caller_max_node_id ]; then
            sg_log 0 "INFO: Response from node \"$LOCAL_HOST_NAME\" to remote node"\
                     "\"`get_node_name_from_nodeid $caller_max_node_id`\" is that, the node"\
                     "\"`get_node_name_from_nodeid $caller_max_node_id`\" is not eligible as"\
                     "the node \"$LOCAL_HOST_NAME\" has lower node id than that of node"\
                     "\"`get_node_name_from_nodeid $caller_max_node_id`\"."
            echo "$FALSE"
            exit 0
        else
            sg_log 0 "INFO: Response from node \"$LOCAL_HOST_NAME\" to remote node"\
                     "\"`get_node_name_from_nodeid $caller_max_node_id`\" is that, the node"\
                     "\"`get_node_name_from_nodeid $caller_max_node_id`\" is eligible as the node"\
                     "\"`get_node_name_from_nodeid $caller_max_node_id`\" has lower node id than"\
                     "that of node \"$LOCAL_HOST_NAME\"."
            echo "$TRUE"
            exit 0
        fi

    # 5. If both are SECONDARY and both are having same seq_num, the lowest node-id will win the race
    elif [ $local_ag_role -eq $SECONDARY ] && [ $caller_role -eq $SECONDARY ]; then
        sg_log 0 "INFO: Both the nodes are SECONDARY."
        get_nodeid_from_nodename $LOCAL_HOST_NAME
        local_node_id=$?
        if [ $local_node_id -lt $caller_max_node_id ]; then
            sg_log 0 "INFO: Response from node \"$LOCAL_HOST_NAME\" to remote node"\
                     "\"`get_node_name_from_nodeid $caller_max_node_id`\" is that, the node"\
                     "\"`get_node_name_from_nodeid $caller_max_node_id`\" is not eligible as"\
                     "the node \"$LOCAL_HOST_NAME\" has lower node id than that of node"\
                     "\"`get_node_name_from_nodeid $caller_max_node_id`\"."
            echo "$FALSE"
            exit 0
        else
            sg_log 0 "INFO: Response from node \"$LOCAL_HOST_NAME\" to remote node"\
                     "\"`get_node_name_from_nodeid $caller_max_node_id`\" is that, the node"\
                     "\"`get_node_name_from_nodeid $caller_max_node_id`\" is eligible as the node"\
                     "\"`get_node_name_from_nodeid $caller_max_node_id`\" has lower node id than"\
                     "that of node \"$LOCAL_HOST_NAME\"."
            echo "$TRUE"
            exit 0
        fi
    fi
fi
