#!/bin/bash

################################################################################
# (C) Copyright 2016-2017 Hewlett Packard Enterprise Development LP
# @(#) Microsoft SQL Server on Linux Monitor script.
# @(#) Product Name                :  HPE Serviceguard extension for Microsoft SQL Server. SGeMSSQL
# @(#) Product Version             :  A.12.20.00
# @(#) Patch Name                  :  
#
# *** Note: This file MUST NOT be edited. ***
#
# Any changes made to this file will be overwritten when you upgrade
# to the next release of Serviceguard Extension for Microsoft SQL Server
#
# Changing this file may lead to unrecoverable package startup
# and/or shutdown problems.
#
################################################################################

RESOURCE_SYSTEM="SQL_Server_Instance_System"
RESOURCE_RESOURCE="SQL_Server_Instance_Resource"
RESOURCE_QUERY_PROC="SQL_Server_Instance_Query_Processing"
RESOURCE_IO_SUBSYS="SQL_Server_Instance_IO_Subsystem"
RESOURCE_EVENTS="SQL_Server_Instance_Events"

# Health states value as warning
typeset -i WARNING=2

MSSQL_HOST_PORT=localhost,$MSSQL_MSSQL_PORT

function process_info
{

    case $1 in
    1|2|3)
        sg_log 1 "INFO: The Value of Instance $2 is = $1"
        cmsetresource -r $3 $1 2> /dev/null
        ;;     
    *)
        sg_log 0 "ERROR: The Value of Instance $2 is = $1"
        sg_log 0 "$output"
        cmsetresource -r $3 4 2> /dev/null
        sg_send_mail "Instance $2 value does not satisy the health criteria for Microsoft SQL Server on Linux. Failing all database instances running on the node `hostname`."
        ;;
    esac
}

################################################################################
# This script checks the Health of Microsoft SQL Server on Linux.
################################################################################
function check_mssql_health
{
    retry=1
    MAX_RETRY=3
   
    while :;
    do
        output=`$SQLCMD -S $MSSQL_HOST_PORT -U $MSSQL_MSSQL_ADMIN -P $MSSQL_MSSQL_PASSWORD -Q "EXEC sp_server_diagnostics" -W`
        retval=$?

        if [[ $retval != 0 ]]
        then 
            if [[ $retry -gt $MAX_RETRY ]]
            then
                sg_log 0 "ERROR: Unable to obtain Health Information of Microsoft SQL Server"
                sg_log 0 "$output"
                return 1;
            else
                sg_log 0 "WARNING: Unable to obtain Health Information of Microsoft SQL Server"
                sg_log 0 "$output"                
                sg_log 0 "Retrying $retry/$MAX_RETRY time"
                (( retry = $retry + 1 ))
                $SLEEP $MSSQL_MONITOR_INTERVAL
                continue
            fi           
        fi

        #The very fact that you are here you are able to get Status so Reset Retry Count
        retry=1

        # if the output is not we expect lets warn and retry in the next monitor interval.
        if `echo "$output" | grep -qw create_time`
        then
            system=`echo "$output" | grep -w "instance system" | awk '{print $5}'`
            resource=`echo "$output" | grep -w "instance resource" | awk '{print $5}'`
            query_processing=`echo "$output" | grep -w "instance query_processing" | awk '{print $5}'`
            io_subsystem=`echo "$output" | grep -w "instance io_subsystem" | awk '{print $5}'`
            events=`echo "$output" | grep -w "instance events" | awk '{print $5}'`

            process_info $system "system" $RESOURCE_SYSTEM
            process_info $resource "resource" $RESOURCE_RESOURCE
            process_info $query_processing "query_processing" $RESOURCE_QUERY_PROC
           
            if [[ $io_subsystem == 1 ]] || [[ $io_subsystem == 2 ]]
            then
                cmsetresource -r $RESOURCE_IO_SUBSYS $io_subsystem 2> /dev/null
            else
                sg_log 0 "ERROR: The Value of Instance io_subsystem is = $io_subsystem"
                sg_send_mail "Instance io_subsystem value does not satisfy the health criteria for Microsoft SQL Server on Linux. Failing all database instances running on the node `hostname`."
                cmsetresource -r $RESOURCE_IO_SUBSYS 3 2> /dev/null
            fi

            if [[ $events == 0 ]]
            then
                cmsetresource -r $RESOURCE_EVENTS -s up 2> /dev/null
            else
                sg_log 0 "ERROR: The Value of Instance events is = $events"
                sg_send_mail "Instance events value does not satisfy the health criteria for Microsoft SQL Server on Linux. Failing all database instances running on the node `hostname`."
                cmsetresource -r $RESOURCE_EVENTS -s down 2> /dev/null
            fi
        else
            sg_log 0 "WARNING: Failed to execute sqlcmd to check sp_server_diagnostics"
        fi

        $SLEEP $MSSQL_MONITOR_INTERVAL
    done
}

function sp_server_diagnostic_halt_monitor
{
    # These value will be accounted for when they restart.
    # This is required because the next start of the package will
    # not cause any issues.

    # As part of going down let us set the 
    # health states to Warning(2)
    cmsetresource -r $RESOURCE_SYSTEM $WARNING 2> /dev/null
    cmsetresource -r $RESOURCE_RESOURCE $WARNING 2> /dev/null
    cmsetresource -r $RESOURCE_QUERY_PROC $WARNING 2> /dev/null
    cmsetresource -r $RESOURCE_IO_SUBSYS $WARNING 2> /dev/null

    # Let us keep it as up
    cmsetresource -r $RESOURCE_EVENTS -s up 2> /dev/null
    exit 0
}

# SIGTERM signal handler to stop the monitor script
trap "sp_server_diagnostic_halt_monitor" SIGTERM

# Start monitoring
#
check_mssql_health
