#!/bin/bash
#############################################################################
# (C) Copyright 2016-2017 Hewlett Packard Enterprise Development LP
# @(#) Micorosoft SQL Server on Linux Utilities
# @(#) Product Name                :  HPE Serviceguard extension for Microsoft SQL Server. SGeMSSQL
# @(#) Product Version             :  A.12.20.00
# @(#) Patch Name                  :  
#
# *** Note: This file MUST NOT be edited. ***
#
# These attributes are used for Microsoft SQL Server on Linux package and
# should not be manually edited. Any changes made to it will
# be overwritten when you upgrade to the next release of Serviceguard Extension
# for Microsoft SQL Server
#
###############################################################################

SESSION_CONTEXT="sp_set_session_context @key = N'external_cluster', @value = N'yes', @read_only = 1;"
typeset query_cmd_retry=3
typeset query_cmd_interval=3
typeset -i clview_retry_times=3
typeset -i query_time_out=60
typeset SQLCMD="sqlcmd -t $query_time_out"
typeset DOC_NAME="HPE Serviceguard Toolkit for MS SQL for Linux User Guide"

# do_cmexec variables
typeset mssql_tmp_dir="/var/tmp"
typeset -i _cmexec_err=255
host_name=$(hostname)
export LOCAL_HOST_NAME=`echo ${host_name%%.*}`

# OS command constants
cmexec=${SGSBIN}/cmexec
cmmodpkg=${SGSBIN}/cmmodpkg
cmrunpkg=${SGSBIN}/cmrunpkg
cmviewcl=${SGSBIN}/cmviewcl
cmcp=${SGSBIN}/cmcp

# Global variable to cache the output of "cmviewcl -v -f line"
typeset VCLVIEW=""

# Global variable to cache the output of "cmviewcl -v -f line -s config"
typeset VCLVIEW_SCONFIG=""

# GR values used in instance_role.mon
# defined in table master.sys.dm_hadr_availability_replica_states
#-----------------------------------------|
# AG State  | MSSQL Values | Servcieguard |
#           |              | Equivalent   |
#           |              | values       |
#-----------|--------------|--------------|
# Resolving |      0       |     1        |
# Primary   |      1       |     2        |
# Secondary |      2       |     3        |
# DOWN/     |     NA       |     4        |
# UNKNOWN   |              |              |
#-----------------------------------------|
# We cannnot reuse the values that are defined by
# MSSQL because we cannot set gr value to zero
typeset -i SG_RESOLVING=1
typeset -i SG_PRIMARY=2
typeset -i SG_SECONDARY=3
typeset -i DOWN=4
typeset -i RESOLVING=0
typeset -i PRIMARY=1
typeset -i SECONDARY=2

# synchronization_health values and description
# defined in table sys.dm_hadr_availability_replica_states
#--------------------------------------------------|
# Synchronization    | MSSQL Values | Servcieguard |
# Health             |              | Equivalent   |
#                    |              | values       |
#--------------------|--------------|--------------|
# NOT_HEALTHY        |      0       |     0        |
# PARTIALLY_HEALTHY  |      1       |     1        |
# HEALTHY            |      2       |     2        |
# UNKNOWN            |      NA      |     NA       |
#--------------------------------------------------|
typeset -i SG_SYNC_NOT_HEALTHY=0
typeset -i SG_SYNC_PARTIALLY_HEALTHY=1
typeset -i SG_SYNC_HEALTHY=2

# Global used for MSSQL login credentials file
typeset CREDENTIALS_FILE="/var/opt/mssql/secrets/.mssql_credentials"

# Global Monitor Defines #
# Defines for role monitor
typeset seq_string="sequence_number"
typeset -i TRUE=0
typeset -i FALSE=1
typeset -i seq_num_from_db
typeset -i role_from_db
typeset -i availability_mode_from_db
typeset -a recover_health_from_db
typeset -a db_is_suspended_from_db
typeset -a db_names_from_db
typeset PRIM_NODE_FROM_FILE
typeset -i SEQ_NUM_FROM_FILE
typeset -i LAG_IN_SECONDS

# Defines for role monitor

###############################################################################
# This function caches cmviewcl -f line output in the variable VCLVIEW.
# Internal Use only
#
#
# Arguments:  None
#
#
# Returns:    none
#
###############################################################################

function cache_clview
{
   typeset -i counter=$clview_retry_times
   sg_log 5 "Function cache_clview"
   sg_log 5 "Refreshing cmviewcl -f line cache"
   VCLVIEW=`$cmviewcl -f line 2>&1`
   if [[ $? -ne 0 ]]; then
      while [[ $counter -gt 0 ]]
      do
         sg_log 0 "Execution of cmviewcl -f line failed"
         sg_log 0 "$VCLVIEW"
         sg_log 0 "Retrying"
         #sleep for retry interval
         sleep $clview_retry_interval
         VCLVIEW=`$cmviewcl -vf line 2>&1`
         if [[ $? -eq 0 ]]; then
            break
         fi
         counter=`expr $counter - 1`
      done
      if [[ $counter -eq 0 ]]; then
         sg_log 0 "Execution of cmviewcl -f line failed"
         sg_log 0 "$VCLVIEW"
         sg_log 0 "Exiting"
         exit 2
      fi
   fi
}

###############################################################################
# This function caches cmviewcl -vf line -s configoutput in the variable
# VCLVIEW_SCONFIG.
# Internal Use only
#
#
# Arguments:  None
#
#
# Returns:    none
#
###############################################################################

function cache_clview_sconfig
{
   typeset -i counter=$clview_retry_times
   sg_log 5 "Function cache_clview_sconfig"
   sg_log 5 "Refreshing cmviewcl -v -f line -s config cache"
   VCLVIEW_SCONFIG=`$cmviewcl -v -f line -s config 2>&1`
   if [[ $? -ne 0 ]]; then
      while [[ $counter -gt 0 ]]
      do
         sg_log 0 "Execution of cmviewcl -v -f line -s config failed"
         sg_log 0 "$VCLVIEW_SCONFIG"
         sg_log 0 "Retrying"
         #sleep for retry interval
         sleep $clview_retry_interval
         VCLVIEW_SCONFIG=`$cmviewcl -v -f line -s config 2>&1`
         if [[ $? -eq 0 ]]; then
            break
         fi
         counter=`expr $counter - 1`
      done
      if [[ $counter -eq 0 ]]; then
         sg_log 0 "Execution of cmviewcl -v -f line -s config failed"
         sg_log 0 "$VCLVIEW_SCONFIG"
         sg_log 0 "Exiting"
         exit 2
      fi
   fi
}

##############################################################################
# Function: detach_dbinstance
# Input Parameter: Database Instance Name
##############################################################################
function detach_dbinstance
{
    sg_log 0 "Detaching Microsoft SQL Database Instance $1:"

    output=`$SQLCMD -U $MSSQL_MSSQL_ADMIN -P $MSSQL_MSSQL_PASSWORD -Q \
           "USE master; ALTER DATABASE $1 SET SINGLE_USER EXEC sp_detach_db \
           @dbname='$1', @keepfulltextindexfile='true';"`
    echo "$output"
}


##############################################################################
# Function: attach_dbinstance
# Input Parameter: Database Instance Name
##############################################################################
function attach_dbinstance
{
    sg_log 0 "Attaching Microsoft SQL Database Instance $1:"

    output=`$SQLCMD -U $MSSQL_MSSQL_ADMIN -P $MSSQL_MSSQL_PASSWORD -Q \
            "USE master; EXEC sp_attach_db $1, '$2', '$3'"`
    echo "$output"
}


##############################################################################
# Function: get_dbinstance
# Input Parameter: Database Instance Name
##############################################################################
function get_dbinstance
{
    output=`$SQLCMD -U $MSSQL_MSSQL_ADMIN -P $MSSQL_MSSQL_PASSWORD -Q \
           "EXEC sp_helpdb N'$1'" -W`
    echo "$output"
}

##############################################################################
# Function: Get all Database which are ONLINE
# Input Parameter: None
##############################################################################
function get_all_online_db
{
    db_output=`$SQLCMD -U $MSSQL_MSSQL_ADMIN -P $MSSQL_MSSQL_PASSWORD -Q \
              "$SESSION_CONTEXT SELECT name, state, state_desc FROM sys.databases \
              WHERE state = 0 and state_desc = 'ONLINE' and name NOT IN \
              ('master', 'tempdb', 'model', 'msdb')" -W -h -1`

    retval=$?
    echo $db_output
}

##############################################################################
# Function: Get all Database which are OFFLINE and State Not Zero
# Input Parameter: None
##############################################################################
function get_all_offline_db
{
    db_output=`$SQLCMD -U $MSSQL_MSSQL_ADMIN -P $MSSQL_MSSQL_PASSWORD -Q \
              "$SESSION_CONTEXT SELECT name, state, state_desc FROM sys.databases \
              WHERE state != 0 and name NOT IN ('master', 'tempdb', 'model', 'msdb')" -W -h -1`

    retval=$?
    echo $db_output
}

##############################################################################
# Function: Get all Database names which are configured
# Input Parameter: None
##############################################################################
function get_all_configured_db
{
    typeset db_output=""
    cmd="$SQLCMD -U $MSSQL_MSSQL_ADMIN -P $MSSQL_MSSQL_PASSWORD -Q \"set nocount on SELECT
         name FROM sys.databases  WHERE name NOT IN ('master', 'tempdb', 'model', 'msdb')\" -W"
    pattern_string="name"
    output=`execute_sql_cmd_on_local_node "$pattern_string" $cmd`
    if [[ $output != "CMD_FAILED" ]]; then
        while read -r line
        do
            if echo "$line" | grep -q -e "^$pattern_string" -e "^----"
            then
                continue
            else
                if [[ "Xdb_output" != "X" ]]; then
                    db_output="$db_output $line"
                else
                    db_output=$line
                fi
            fi
        done  <<<"$output"
    fi

    echo $db_output
}

##############################################################################
# Function: Enable Always on Availability Groups
# Input Parameter: None
##############################################################################
function enable_aoai
{
    sg_log 0 "Enabling Always on Availability Groups"

    /opt/mssql/bin/mssql-conf set hadrenabled 1 > /dev/null
}

##############################################################################
# Function: Check if Always on Availability Groups is Enabled
# Input Parameter: None
# Returns: 0 If TRUE and 1 if FALSE.
##############################################################################
function is_aoai_enabled
{
    typeset -i hadrenbaled=0
    if [[ -f /var/opt/mssql/mssql.conf ]]; then
        hadrenbaled=`cat /var/opt/mssql/mssql.conf | grep hadrenabled | awk '{print $3}'`
    else
        echo "ERROR: /var/opt/mssql/mssql.conf does not exists"
        return 1
    fi

    if (( hadrenbaled == 1 ))
    then
        # Return TRUE
        return 0
    fi
    # Return FALSE
    return 1
}

##############################################################################
# Function: Join a given Secondary Replica to AG if not Joined already
# Input Parameter: AG Name
##############################################################################
function join_secondary_replica_to_ag
{
    AG=$1;
    sg_log 0 "Setting the role of Availability Group $AG to Secondary"

    output=`$SQLCMD -U $MSSQL_MSSQL_ADMIN -P $MSSQL_MSSQL_PASSWORD -Q \
           "$SESSION_CONTEXT IF NOT EXISTS(SELECT * FROM sys.availability_groups WHERE name = '$AG') \
            ALTER AVAILABILITY GROUP [$AG] JOIN WITH (CLUSTER_TYPE = EXTERNAL)"`

    echo $output
}

##############################################################################
# Function: failover_ag
# Input Parameter: Failover an Availability Group.
##############################################################################
function failover_ag
{
    AG=$1;
    sg_log 0 "FAILOVER: Setting the role of Availability Group $AG to Primary"

    output=`$SQLCMD -U $MSSQL_MSSQL_ADMIN -P $MSSQL_MSSQL_PASSWORD -Q \
           "$SESSION_CONTEXT ALTER AVAILABILITY GROUP [$AG] FAILOVER" -W`

    echo "$output"
}

##############################################################################
# Function: set_role_to_secondary
# Input Parameter: Availability Group Name
##############################################################################
function set_role_to_secondary
{
    AG=$1;
    sg_log 0 "Setting the role of Availability Group $AG to Secondary"

    output=`$SQLCMD -U $MSSQL_MSSQL_ADMIN -P $MSSQL_MSSQL_PASSWORD -Q \
           "$SESSION_CONTEXT ALTER AVAILABILITY GROUP [$AG] SET (ROLE = SECONDARY)" -W`

    echo "$output"
}

##############################################################################
# Function: check_if_local_node_is_primary_replica
# Input Parameter: Availability Group Name
##############################################################################
function check_if_local_node_is_primary_replica
{
    $SQLCMD -U $MSSQL_MSSQL_ADMIN -P $MSSQL_MSSQL_PASSWORD -Q "set nocount on SELECT * \
    from master.sys.dm_hadr_availability_replica_states;" -W -h -1 | grep -q PRIMARY
    if (( $? != 0 )); then
        return 1
    fi
    return 0
}

##############################################################################
# Function: get_mssql_server_package_name
# Input: NONE
# Output: The name of the package which has mssqlserever module configured
#         Else empty string
##############################################################################
function get_mssql_server_package_name
{
    typeset PKG_NAME=""
    PKG_NAME=`$cmviewcl -vf line -s config | \
              grep "module_name:sgemssql/mssqlserver|module_name=sgemssql/mssqlserver"\
              | awk -F"[:|=]" '{print $2}'`
    echo "$PKG_NAME"
}

########################################################################
# This function is used to validate the login credentials
# for AOAI MNP packages.
########################################################################
function validate_login_credentials_for_package
{
   typeset _mssql_admin=$1
   typeset _mssql_password=$2
   typeset _node_name=$3
   typeset -i _ret
   validate_login_credentials $_mssql_admin $_mssql_password $_node_name
   _ret=$?

   if (( $_ret == 1));
   then
       echo "ERROR: Invalid Credentials for the user $_mssql_admin. Provide valid Username and Password."
       return 1
   fi
   return 0
}

########################################################################
# This function is used to distrubute the login credentials file
# {/var/opt/mssql/secrets/.mssql_credentials} to other nodes on which
# AG is configured.
########################################################################
function distribute_login_credentials_file
{
    file=$1
    shift 1
    nodes=$*

    for node in $nodes
    do
       if [[ "$node" != "$LOCAL_HOST_NAME" ]];
       then
            $cmcp $file $node:$file 2>/dev/null
       fi
    done
}

########################################################################
# This function is used to retrieve username and password from
# {/var/opt/mssql/secrets/.mssql_credentials}
########################################################################

function get_user_name_and_password_from_login_credentials_file
{
    typeset _user_str
    typeset _passwd_str
    typeset -i ret=0

    if [ -f $CREDENTIALS_FILE ]
    then
        creds=`cat $CREDENTIALS_FILE | grep -v "^#"`
        _user_str=`echo $creds | awk '{print $1}'`
       _passwd_str=`echo $creds | awk '{print $2}'`

        if [[ $_user_str != "" ]] && [[ $_passwd_str != "" ]];
        then
            MSSQL_MSSQL_ADMIN=$_user_str
            MSSQL_MSSQL_PASSWORD=$_passwd_str
            export MSSQL_MSSQL_ADMIN
            export MSSQL_MSSQL_PASSWORD
        fi
    else
        ret=1
    fi

    return $ret
}

########################################################################
# This function is used to validate the username and password from
# {/var/opt/mssql/secrets/.mssql_credentials}
########################################################################

function validate_login_credentials
{
    typeset _mssql_admin=$1
    typeset _mssql_password=$2
    typeset _node_name=$3
    typeset _local_output=$mssql_tmp_dir/cmexec_validate_cred_$$
    typeset -i _ret=0

    # required for do_cmexec
    typeset _output=$mssql_tmp_dir/cmexec.${0##*/}.login_credentials_$$
    typeset _stderr=$mssql_tmp_dir/cmexec.err.${0##*/}.login_credentials_$$

    set_retry_interval

    cmd="$SQLCMD"
    args="-U $_mssql_admin -P $_mssql_password -Q ''"

    cmd_opts="$_node_name $_output $_stderr"
    do_cmexec $cmd_opts $cmd $args  >$_local_output 2>&1
    rt=$?
    if [ $rt -gt 0 ]; then
         grep -q "ERROR" $_local_output
         _ret=$?
    fi

    rm -f $_local_output
    return $_ret
}

##############################################################################
# Function: get_mssql_tcp_port
# Output: configured port
##############################################################################
function get_mssql_tcp_port
{
    default_port="1433"
    conf_file="/var/opt/mssql/mssql.conf"
    if [[ -e $conf_file ]]
    then
        port=`awk '/tcpport =/ {print $3}' $conf_file`
        if [[ -z "$port" ]]; then
            echo $default_port
        else
            echo $port
        fi
    fi
}

###############################################################################
# Function: validate_mssql_server_configuration
#
# Validate if the Microsoft SQL Server on Linux package is installed or not.
#
###############################################################################
function validate_mssql_server_configuration
{
    sg_log 5 "Microsoft SQL Server on Linux Validation"

    # Validate if Microsoft SQL Server is installed
    rpm -qi mssql-server > /dev/null
    status=$?
    if [[ $status -ne 0 ]]; then
        sg_log 0 "ERROR: Microsoft SQL Server on Linux is not installed"
        return 1
    fi


    return 0
}



###############################################################################
# Function: validate_mssql_port
#
# Validate if the provided port is valid.
#
###############################################################################
function validate_mssql_port
{
    PORT=`get_mssql_tcp_port`
    if [[ "$PORT" == "$MSSQL_MSSQL_PORT" ]]
    then
        return 0
    fi
    sg_log 0 "ERROR: Invalid Microsoft SQL Server Port"
    sg_log 0 "ERROR: The port given as part of package configuration does not match"\
             "with what is configured."
    return 1
}

###############################################################################
# Function: does_mssql_server_package_exist
#
# validate if MSSQL Server package is already configured
#
###############################################################################
function does_mssql_server_package_exist
{
    PKG_NAME=`get_mssql_server_package_name`

    #If Online configuration return 0
    if [[ -n "$PKG_NAME" ]] && [[ "$PKG_NAME" != "$SG_PACKAGE_NAME" ]]
    then
       sg_log 0 "ERROR: Microsoft SQL Server on Linux package is already configured under the"\
                "package name $PKG_NAME"
       sg_log 0 "ERROR: Only one instance of Microsoft SQL Server on Linux can be configured in"\
                "a Serviceguard cluster."
       return 1
    fi

    return 0
}

###############################################################################
# Function: check_if_ag_already_configured_in_a_package
#
# validate if AOAI package already configured with the provided AG.
#
###############################################################################
function check_if_ag_already_configured_in_a_package
{
    typeset AG_PKG_NAME
    AG_PKG_NAME=`echo "$VCLVIEW_SCONFIG" | grep -w "sgemssql/mssqlaoai/availability_group=$MSSQL_AVAILABILITY_GROUP" \
                 | awk -F"[:|=]" '{print $2}'`
    #If Online configuration return 0
    if [[ -z $AG_PKG_NAME ]] || [[ "$AG_PKG_NAME" == "$SG_PACKAGE_NAME" ]]
    then
        return 0
    fi

    sg_log 0 "ERROR: Package $AG_PKG_NAME already exists with the Availability Group $MSSQL_AVAILABILITY_GROUP"
    return 1
}

#######################################################################
# Function: is_package_up
#
# checks if the package is up or not
#######################################################################

function is_package_up
{
    typeset pkgname=$1
    typeset status

    # Refresh Cache
    cache_clview

    status=`echo "$VCLVIEW" | awk -F "[:|=]" '(($1 == "package") && \
                ($2 == "'$pkgname'") && ($3 == "status")) {print $4}'`

    if [[ "$status" = "up" ]]
    then
        return 0
    else
        return 1
    fi
}

###############################################################################
# Function: validate_aoai_parameters
#
# validate AOAI package parameters if MSSQL is Server is UP.
#
###############################################################################
function validate_aoai_parameters
{
    typeset VALIDATE_AG_DB_NAME_ARRAY
    typeset VALIDATE_AG_NODE_NAME_ARRAY
    typeset VALIDATE_AVAILABILITY_MODE_ARRAY
    typeset CLUSTER_NODES
    typeset -i retVal=0
    typeset -i isPrimaryNode=1

    if pgrep -x "sqlservr" > /dev/null
    then
        get_user_name_and_password_from_login_credentials_file
        if [ $? -ne 0 ]; then
            sg_log 0 "ERROR: MSSQL Credential validation failed on the"\
                     "node \"$LOCAL_HOST_NAME\""
            exit 1
        fi

        if [[ -n "$MSSQL_MSSQL_ADMIN" ]] && [[ -n "$MSSQL_MSSQL_PASSWORD" ]]
        then
            output=`$SQLCMD -U $MSSQL_MSSQL_ADMIN -P $MSSQL_MSSQL_PASSWORD -Q "set nocount on SELECT \
            ag.required_synchronized_secondaries_to_commit, d.database_name, r.replica_server_name, \
            r.availability_mode_desc from master.sys.availability_groups ag inner \
            join master.sys.availability_replicas r on ag.group_id = r.group_id inner join \
            master.sys.availability_databases_cluster d on ag.group_id = d.group_id inner join \
            master.sys.dm_hadr_availability_replica_states m on r.replica_id = m.replica_id WHERE \
            ag.name = '$MSSQL_AVAILABILITY_GROUP';" -W -h -1`

            if [[ -z $output ]]
            then
                sg_log 0 "ERROR: Availability Group $MSSQL_AVAILABILITY_GROUP does not exists."
                return 1
            fi

            while read -r line
            do
                VALIDATE_REQUIRED_SYNCHRONIZED_SECONDARIES_TO_COMMIT=`echo $line | awk '{print $1}'`
                TEMP_DB_NAME=(`echo $line | awk '{print $2}'`)
                if ! `echo "${VALIDATE_AG_DB_NAME_ARRAY[@]}" | grep -qw $TEMP_DB_NAME`
                then
                    VALIDATE_AG_DB_NAME_ARRAY+=("$TEMP_DB_NAME")
                fi
                TEMP_NODE_NAME=(`echo $line | awk '{print $3}'`)
                if ! `echo "${VALIDATE_AG_NODE_NAME_ARRAY[@]}" | grep -qw $TEMP_NODE_NAME`
                then
                    VALIDATE_AG_NODE_NAME_ARRAY+=("$TEMP_NODE_NAME")
                    VALIDATE_AVAILABILITY_MODE_ARRAY+=(`echo $line | awk '{print $4}' | cut -d'_' -f 1`)
                fi
            done <<<"$output"
            
            # Only the primary node has more than 1 entry in the output
            # Thus, this code must get executed only on the primary replica node
            # of a particular Availability Group (AG)
            if [[ ${#VALIDATE_AVAILABILITY_MODE_ARRAY[@]} -gt 1 ]]
            then
                NO_SYNCH_NODES=0
                for I in ${VALIDATE_AVAILABILITY_MODE_ARRAY[@]}
                do
                    if [[ "$I" == "SYNCHRONOUS" ]]
                    then
                        (( NO_SYNCH_NODES = NO_SYNCH_NODES + 1 ))
                    fi
                done
                
                if [[ $NO_SYNCH_NODES -lt 2 ]]
                then
                    sg_log 0 "ERROR: Availability Group $MSSQL_AVAILABILITY_GROUP is configured with only one replica."\
                    "A minimum of two SYNCHRONOUS_COMMIT replicas must be configured."
                    return 1
                fi
            fi

            if [[ ${SG_NODE_NAME[0]} == "*" ]]
            then
                CLUSTER_NODES=`echo "$VCLVIEW_SCONFIG" | awk -F"[:|=]" '(($1 == "node") && ($3 == "name")){print $4}'`
            else
                CLUSTER_NODES=${SG_NODE_NAME[@]}
            fi

            #validate REQUIRED_SYNCH_SECONDARIES_TO_COMMIT
            if [[ $VALIDATE_REQUIRED_SYNCHRONIZED_SECONDARIES_TO_COMMIT != ${MSSQL_REQUIRED_SYNCHRONIZED_SECONDARIES_TO_COMMIT} ]]
            then
                sg_log 0 "ERROR: Invalid MSSQL_REQUIRED_SYNCHRONIZED_SECONDARIES_TO_COMMIT"
                sg_log 0 "ERROR: The MSSQL_REQUIRED_SYNCHRONIZED_SECONDARIES_TO_COMMIT given as part of"\
                         "package configuration does not match with what is configured."
                retVal=1
            fi

            #validate Nodes. If Node is Primary Replica Validation will happen for all nodes.
            for I in ${VALIDATE_AG_NODE_NAME_ARRAY[@]}
            do
                if ! `echo "${CLUSTER_NODES[@]}" | grep -qw $I`
                then
                    sg_log 0 "ERROR: Availability Group Replica $I is not part of the package configuration."
                    retVal=1
                fi
            done

            #Validate if all DBs in AG is part of package configuration.
            for I in ${VALIDATE_AG_DB_NAME_ARRAY[@]}
            do
                if ! `echo "${MSSQL_AOAI_DATABASE[@]}" | grep -qw $I`
                then
                    sg_log 0 "ERROR: Database $I is not part of the package configuration."
                    retVal=1
                fi
            done

            #Validate if the DBs in package configuration is part of AG.
            for J in ${MSSQL_AOAI_DATABASE[@]}
            do
                if ! `echo "${VALIDATE_AG_DB_NAME_ARRAY[@]}" | grep -qw $J`
                then
                    sg_log 0 "ERROR: Database $J is not part of the $MSSQL_AVAILABILITY_GROUP."
                    retVal=1
                fi
            done

            #Validate AVAILABILITY_MODE if nodes are valid
            index_v=0
            index_c=0

            for i in ${VALIDATE_AG_NODE_NAME_ARRAY[@]}
            do
               for j in ${CLUSTER_NODES[@]}
               do
                  if [ $j = $i ]
                  then
                      if [[ ${VALIDATE_AVAILABILITY_MODE_ARRAY[index_v]} != ${MSSQL_AVAILABILITY_MODE[index_c]} ]];
                      then
                          sg_log 0 "ERROR: The Availability Mode given for the node $i as part of"\
                                   "package configuration does not match with what is configured."
                          retVal=1
                      fi
                      break
                  fi
                  (( index_c = index_c + 1 ))
              done
              index_c=0
              (( index_v = index_v + 1 ))
            done
        fi
    else
        # Validate Number of synchronous replicas from the configuration file
        # if mssql server is not running
        NO_OF_SYNCH_NODES=0
        for I in ${MSSQL_AVAILABILITY_MODE[@]}
        do
            if [[ "$I" == "SYNCHRONOUS" ]]
            then
                (( NO_OF_SYNCH_NODES = NO_OF_SYNCH_NODES + 1 ))
            fi
        done
        
        if [[ $NO_OF_SYNCH_NODES -lt 2 ]]
        then
            sg_log 0 "ERROR: Availability Group $MSSQL_AVAILABILITY_GROUP is configured with only one replica."\
            "A minimum of two SYNCHRONOUS_COMMIT replicas must be configured."
            return 1
        fi
    fi

    return $retVal
}

##############################################################################
# Function: get_the_role_of_the_ag_on_local_replica
# Uses the sqlcmd to query the following SYS tables in the DB,
# DB Name: sys.availability_groups
# DB Name: sys.dm_hadr_database_replica_states
# And gets the role of the local replica from
# sys.dm_hadr_database_replica_states for matching group_id with
# sys.availability_groups table.
# Input: AG name
# Output: Role of the local replica.
##############################################################################
function get_the_role_of_the_ag_on_local_replica
{
    typeset ag_name=$1
    typeset -i role
    typeset counter=$query_cmd_retry
    typeset -i i=1

    role=`$SQLCMD -U $MSSQL_MSSQL_ADMIN -P $MSSQL_MSSQL_PASSWORD -Q  "set nocount on
         SELECT ars.role FROM sys.availability_groups ag,
         master.sys.dm_hadr_availability_replica_states ars
         WHERE ag.name = '$ag_name' AND ars.is_local = 1 AND ag.group_id = ars.group_id;" -W -h -1`

    if [[ "X$role" != "X" ]]; then
        return $role
    else
        # Failed to execute the $SQLCMD, lets retry before giving up
        sg_log 3 "ERROR: sqlcmd Failed:"
        sg_log 3 "SELECT ars.role FROM sys.availability_groups ag, "\
                 "master.sys.dm_hadr_availability_replica_states ars"\
                 " WHERE ag.name = '$ag_name' AND ars.is_local = 1 AND ag.group_id = ars.group_id;"
        sg_log 0 "ERROR: Failed to get the role of the Availability Group \"$ag_name\" on the node."\
                 " $LOCAL_HOST_NAME."
        while [[ $counter -gt 0 ]]
        do
            sg_log 0 "INFO: Retrying the sqlcmd command: $i/3"
            role=`$SQLCMD -U $MSSQL_MSSQL_ADMIN -P $MSSQL_MSSQL_PASSWORD -Q  "set nocount on
                 SELECT ars.role FROM sys.availability_groups ag,
                 master.sys.dm_hadr_availability_replica_states ars
                 WHERE ag.name = '$ag_name' AND ars.is_local = 1
                 AND ag.group_id = ars.group_id;" -W -h -1`
                 if [[ "X$role" != "X" ]]; then
                    break
                fi
            (( counter = counter - 1 ))
            (( i = i + 1 ))
            sleep $query_cmd_interval
        done

        if [[ $counter -eq 0 ]]; then
            # Giving up, could not get the info even after 3 retries.
            sg_log 0 "ERROR: Failed to execute the sqlcmd, setting the role of the Availability"\
            "Group \"$ag_name\" on node $LOCAL_HOST_NAME as UNKNOWN"
            return $DOWN
        else
            return $role
        fi
    fi
}

##############################################################################
# Function: map_sql_server_ag_role_to_sg_ag_role
# Input: role value from the DB
# Output: Equivalent mapped value in Serviceguard
##############################################################################
function map_sql_server_ag_role_to_sg_ag_role
{
    typeset -i role=$1
    case $role in
        0) return $SG_RESOLVING ;;
        1) return $SG_PRIMARY ;;
        2) return $SG_SECONDARY ;;
        *) return $DOWN ;;
    esac
}

##############################################################################
# Function: get_role_description_from_gr
# Input: GR value of instance monitor
# Output: String value for the GR value
##############################################################################
function get_role_description_from_gr
{
    typeset -i role=$1
    case $role in
        0) echo "DEFAULT" ;;
        1) echo "RESOLVING" ;;
        2) echo "PRIMARY" ;;
        3) echo "SECONDARY" ;;
        *) echo "UNKNOWN" ;;
    esac
}

##############################################################################
# Function: check_if_localhost_is_detached
# Input: None
# Returns: return value 0 if grep is successfull else returns value 1. 
##############################################################################
function check_if_localhost_is_detached
{
    AG_PKG_NAME=$1
    local_host=$2 

    $cmviewcl -f line -p $AG_PKG_NAME | awk '/^node:[^|]*\|status=detached/' | \
        grep -qw $local_host 
    ret_val=$?
    return $ret_val 
}


##############################################################################
# Function: set_gr_value_on_local_node
# Input: Name of the GR, New value
##############################################################################
function set_gr_role_on_local_node
{
    typeset gr_name=$1
    typeset -i new_gr_value=$2
    typeset AG=$3
    typeset -i current_gr_value=0
    ag_pkg_name=`get_ag_pkg_name_from_ag_name $AG`


    # check whether current_gr_value is 0 or not as it will be not avaiable when package is detached.
    # we are checking for 0 here instead of null as current_gr_value is typeset to integer.
    check_if_localhost_is_detached $ag_pkg_name $LOCAL_HOST_NAME
    ret_val=$? 
    if [[ $ret_val -ne 0 ]]; then
        current_gr_value=`$SGSBIN/cmgetresource -r $gr_name 2> /dev/null | grep $gr_name | awk '{print $2}'`
        if [[ $current_gr_value -eq $new_gr_value ]]; then
            # Same as before just exit
            return 0
        fi

        $SGSBIN/cmsetresource -r $gr_name $new_gr_value 2> /dev/null
        sg_log 0 "The Availability Group:\"$AG\" role on node $LOCAL_HOST_NAME"\
                 "has changed from `get_role_description_from_gr $current_gr_value` to"\
                 "`get_role_description_from_gr $new_gr_value`"
    fi
}

##################################################################################
# Function: Modify the value of REQUIRED_SYNCHRONIZED_SECONDARIES_TO_COMMIT
# Input: AG_NAME, New value of REQUIRED_SYNCHRONIZED_SECONDARIES_TO_COMMIT, Primary Replica node name
# This command has to be executed only on the primary replica.
##################################################################################
function update_req_sync_secondary_to_commit
{
    typeset AG_NAME=$1
    typeset new_commit=$2
    typeset master_node=$3

    # required for do_cmexec
    typeset _output=$mssql_tmp_dir/cmexec.${0##*/}.deploy_"$AG_NAME"
    typeset _stderr=$mssql_tmp_dir/cmexec.err.${0##*/}.deploy_"$AG_NAME"

    cmd="$SQLCMD"
    args="-U $MSSQL_MSSQL_ADMIN -P $MSSQL_MSSQL_PASSWORD -Q \"ALTER AVAILABILITY GROUP
          $AG_NAME SET (REQUIRED_SYNCHRONIZED_SECONDARIES_TO_COMMIT=$new_commit);\""

    cmd_opts="$master_node $_output $_stderr"
    output=`do_cmexec $cmd_opts $cmd $args`
}

##################################################################################
# Function: Modify the value of AVAILABILITY_MODE
# Input: AG_NAME, NODE_NAME, New value of AVAILABILITY_MODE, Primary Replica node name
# This command has to be executed only on the primary replica.
##################################################################################
function update_availability_mode_for_the_ag
{
    typeset AG_NAME=$1
    typeset NODE_NAME=$2
    typeset new_avail_mode=$3
    typeset master_node=$4

    # required for do_cmexec
    typeset _output=$mssql_tmp_dir/cmexec.${0##*/}.deploy_"$AG_NAME"
    typeset _stderr=$mssql_tmp_dir/cmexec.err.${0##*/}.deploy_"$AG_NAME"

    cmd="$SQLCMD"
    args="-U $MSSQL_MSSQL_ADMIN -P $MSSQL_MSSQL_PASSWORD -Q \"ALTER AVAILABILITY GROUP
         $AG_NAME MODIFY REPLICA ON '$NODE_NAME' WITH (AVAILABILITY_MODE = $new_avail_mode);\""

    cmd_opts="$master_node $_output $_stderr"
    output=`do_cmexec $cmd_opts $cmd $args`
}

function get_node_name_from_nodeid
{
    node_id=$1
    nname=`$cmviewcl -f line -s config\
           | awk '/^node:[^|]*\|id='$node_id'/' | sed 's/|id='${node_id}'//;s/node://'`
    echo $nname
}

##################################################################################
# Function: Modify the value of retrieve_and_distribute_information
# Input: AG_NAME, foot_print file
# This function has to be called only by Primary Replica. Its callers responsibility to ensure that.
##################################################################################

function retrieve_and_distribute_information
{
    AG=$1
    file=$2
    AG_PKG_NAME=$3
    args="$SQLCMD -U $MSSQL_MSSQL_ADMIN -P $MSSQL_MSSQL_PASSWORD -Q \"set nocount on SELECT \
              ag.$seq_string, \
              drs.secondary_lag_seconds, \
              adc.database_name, \
              ar.replica_server_name, \
              ars.synchronization_health \
          FROM sys.dm_hadr_database_replica_states drs, sys.availability_groups ag, \
          master.sys.availability_replicas ar, master.sys.dm_hadr_availability_replica_states ars, \
          master.sys.availability_databases_cluster adc \
          WHERE \
              ag.name = '$AG' AND \
              drs.group_id = ag.group_id AND \
              drs.group_database_id = adc.group_database_id AND \
              ar.replica_id = drs.replica_id AND \
              ar.replica_id = ars.replica_id;\" -W"
    output=`execute_sql_cmd_on_local_node "$seq_string" $args`
    if [[ "$output" != "CMD_FAILED" ]]; then
        if [[ -s $file ]]; then
            rm -f $file
        fi
        while read -r line
        do
            if echo "$line" | grep -q -e "^${seq_string}" -e "--*"
            then
                continue
            else
                echo "$line" >> $file
            fi
        done <<< "$output"
        # Append the last line of the file with the primary node name so that
        # other nodes will get to know who is primary.
        echo "primary_node:$LOCAL_HOST_NAME" >> $file
        up_nodes=`get_pkg_up_and_detached_nodes $AG_PKG_NAME`
        for node in $up_nodes
        do
            # We need to find all the nodes which might reachable.
            # UP_NODES are the nodes which have AG MNP packages running.
            if [[ "$node" != "$LOCAL_HOST_NAME" ]];then
                $cmcp $file $node:$file
            fi
        done
    else
        sg_log 0 "ERROR: No write was done to the file $file as sqlcmd failed to execute"
       return
    fi

}


##############################################################################
# Function: execute_sql_cmd_on_local_node
# Input Parameter:
#   SQL cmd string
#   pattern to be verified if the command has succeeded
# Output:
#   The SQL command output
##############################################################################
function execute_sql_cmd_on_local_node
{
    pattern_to_check=$1
    shift 1
    cmd=$*
    typeset counter=$query_cmd_retry
    typeset -i i=1

    output=`eval $cmd`
    echo "$output" | head -1 | grep -qw $pattern_to_check
    if (( $? != 0 ))
    then
        # Failed to execute the sqlcmd, lets retry before giving up
        sg_log 0 "ERROR: Failed to execute sqlcmd on $LOCAL_HOST_NAME:"
        sg_log 3 "$cmd"
        get_user_name_and_password_from_login_credentials_file
        if [ $? -ne 0 ]; then
            # It cannot come to this state, we have had until now but suddenly went missing
            sg_log 0 "ERROR: MSSQL Credential validation failed on the"\
                     "node \"$LOCAL_HOST_NAME\""
            exit 1
        fi

        while [[ $counter -gt 0 ]]
        do
            sg_log 0 "INFO: Retrying the sqlcmd command: $i/3"
            output=`eval $cmd`
            echo "$output" | grep -qw $pattern_to_check
            if (( $? != 0 )); then
                break
            fi
            (( counter = counter - 1 ))
            (( i = i + 1 ))
            sleep $query_cmd_interval
        done

    fi
    if [[ $counter -eq 0 ]]; then
        # Giving up, could not get the info even after 3 retries.
        sg_log 0 "ERROR: Failed to execute the sqlcmd, "
        echo "CMD_FAILED"
    else
        echo "$output"
    fi
}

##############################################################################
# Function: get_sync_health_descriptor
# Input: integer value of the synchronization_health
# Output: Equivalent string value
##############################################################################
function get_sync_health_descriptor
{
    typeset -i role=$1
    case $role in
        $SG_SYNC_NOT_HEALTHY ) echo "NOT_HEALTHY" ;;
        $SG_SYNC_PARTIALLY_HEALTHY ) echo "PARTIALLY_HEALTHY" ;;
        $SG_SYNC_HEALTHY ) echo "HEALTHY" ;;
        *) echo "UNKNOWN" ;;
    esac
}

##############################################################################
# Function: get_synchronization_health_from_file
# Input Parameter:
#   file_name
#   node name
# Globals
#   db_names_from_db
# Output:
#   TRUE if the state is HEALTHY(2)
#   FALSE if the state is NOT_HEALTHY(0) or PARTIALLY_HEALTHY(1)
# Note: The caller has to ensure that the footprint file exists before calling 
# this function.
##############################################################################
function get_synchronization_health_from_file
{
    # Check if DB is not suspended. We cannot promote to PRIMARY if DB is suspended
    # on this node.
    typeset file=$1
    typeset _node_name=$2
    typeset local_node_values_from_file=""

    typeset -i R=0
    typeset -i ret=$TRUE

    local_node_values_from_file=`cat $file | grep -v primary_node | grep $_node_name`

    # Now the local_node_values_from_file will have the information of each db on local node.
    # We need to get the synchronization_health for each DB.
    while (($R < ${#db_names_from_db[*]}))
    do
        sync_health=`echo "$local_node_values_from_file" | grep ${db_names_from_db[$R]} |\
                     awk '{print $5}'`
        if [ $sync_health -ne $SG_SYNC_HEALTHY ]; then
            ret=$FALSE
            sg_log 0 "INFO: For Availability Group:\"$AG\" on node \"$_node_name\", the database"\
                    "\"${db_names_from_db[$R]}\" synchronization is"\
                    "`get_sync_health_descriptor $sync_health`."
        fi
        (( R = R + 1 ))
    done
    return $ret
}

##############################################################################
# Function: get_commit_mode_from_config
# Input Parameter:
#   node_name
# GLOBALS:
#   cmgetpkgenv variables
#       MSSQL_NODENAME
#       MSSQL_AVAILABILITY_MODE
# Output:
#   TRUE if it is ASYNC commit mode
#   FALSE if it is SYNC commit mode
# This function cannot be used as part of cmdeploymssql since we use
# cmgetpkgenv variables
##############################################################################
function get_commit_mode_from_config
{
    _node_name=$1
    typeset -i R=0
    typeset -i ret=$FALSE
    while (($R < ${#MSSQL_NODENAME[*]}))
    do
        if [[ "$_node_name" == "${MSSQL_NODENAME[$R]}" ]]; then
            if [[ "ASYNCHRONOUS" == "${MSSQL_AVAILABILITY_MODE[$R]}" ]]; then
                ret=$TRUE
                break
            fi
        fi
        (( R = R + 1 ))
    done
    return $ret
}

##############################################################################
# Function: are_all_db_in_healthy_state_for_a_node
# Input Parameter:
#   AG
#   node name
# Globals
#   db_names_from_db
#   db_is_suspended_from_db
# Output:
#   TRUE if there is a LAG
#   FALSE if there is no LAG
##############################################################################
function are_all_db_in_healthy_state_for_a_node
{
    # Check if DB is not suspended. We cannot promote to PRIMARY if DB is suspended
    # on this node.
    typeset AG=$1
    typeset _node_name=$2

    typeset -i R=0
    typeset -i ret=$TRUE

    while (($R < ${#db_names_from_db[*]}))
    do
        if [ ${db_is_suspended_from_db[$R]} -ne 0 ]; then
            ret=$FALSE
            sg_log 0 "INFO: For Availability Group:\"$AG\" on node \"$_node_name\", the database"\
                     "\"${db_names_from_db[$R]}\" is suspended."
        fi
        (( R = R + 1 ))
    done
    return $ret
}

function get_nodeid_from_nodename
{
    _nn=$1
    nodeid=`$cmviewcl -vf line -s config | awk -F= '/^node:'$_nn'\|id=/ { print $2 }'`
    return $nodeid
}

##############################################################################
# Function: get_list_of_all_up_nodes_in_sync_state
# Input Parameter:
#   up node list
# Globals
#   get_commit_mode_from_config
# Output:
#   Node list which are SYNC
##############################################################################
function get_list_of_all_up_nodes_in_sync_state
{
    up_nodes=$*

    SYNC=$FALSE
    typeset _sync_up_nodes
    for n in $up_nodes
    do
        get_commit_mode_from_config $n
        commit_mode=$?
        if [ $commit_mode -eq $SYNC ]; then
            if [[ "X$_sync_up_nodes" == "X" ]]; then
                _sync_up_nodes=$n
            else
                _sync_up_nodes="$_sync_up_nodes $n"
            fi
        fi
    done

    echo "$_sync_up_nodes"
}

##############################################################################
# Function: get_seq_num_from_file
# Input Parameter:
#   NONE
# Globals
#   SEQ_NUM_FROM_FILE
#   PRIM_NODE_FROM_FILE
#   get_primary_node_name_from_file
# Output:
##############################################################################
function get_seq_num_from_file
{
    file=$1

    if [ -f $file ]; then
        get_primary_node_name_from_file $file
        SEQ_NUM_FROM_FILE=`cat $file | grep -w $PRIM_NODE_FROM_FILE | head -1 |  awk '{print $1}'`
    else
        SEQ_NUM_FROM_FILE=0
    fi
}

##############################################################################
# Function: get_primary_node_name_from_file
# Input Parameter:
#   NONE
# Globals
#   PRIM_NODE_FROM_FILE
# Output:
##############################################################################
function get_primary_node_name_from_file
{
    file=$1

    PRIM_NODE_FROM_FILE=`cat $file | grep -w primary_node | cut -d ':' -f2`
}

##############################################################################
# Function: get_lag_in_seconds_from_file
# Input Parameter:
#   foot_print_file
#   node name
# Globals
#   LAG_IN_SECONDS
# Output:
#   LAG_IN_SECONDS
# This function captures the highest LAG for a given node.
# There will be multiple lines for each node for each DB in the AG
##############################################################################
function get_lag_in_seconds_from_file
{
    file=$1
    _node_name=$2

    if [ -f $file ]; then
        _lags=`cat $file | grep -w $_node_name | awk '{print $2}'`
        for lag in $_lags
        do
            if [[ "$lag" == "0" ]] || [[ "$lag" == "NULL" ]]; then
                continue
            else
                if [[ "X$LAG_IN_SECONDS" == "X" ]] || [[ $LAG_IN_SECONDS -lt $lag ]]; then
                    LAG_IN_SECONDS=$lag
                fi
            fi
        done
    fi
}

##############################################################################
# Function: do_i_lag_in_sec_from_file
# Input Parameter:
#   foor_print_file
#   node name
# GLOBALS
#   foot_print_file
# Output:
#   TRUE if there is a LAG
#   FALSE if there is no LAG
##############################################################################
function do_i_lag_in_sec_from_file
{
    file=$1
    _node_name=$2

    unset LAG_IN_SECONDS
    get_lag_in_seconds_from_file $file $_node_name
    if [[ "X$LAG_IN_SECONDS" != "X" ]] && [[ $LAG_IN_SECONDS -gt 0 ]]; then
        return $TRUE
    fi
    return $FALSE
}

function read_the_file_params_for_prev_primary_and_sequence_num
{
    file=$1

    if [ -f $file ]; then
        get_primary_node_name_from_file $file
        get_seq_num_from_file $file
    else
        return 0
    fi

}

function remove_start_file
{
    start_file=$1
    if [[ -f $start_file ]]; then
        rm -f $start_file
    fi
}

##############################################################################
# Function: can_i_be_the_master
# Input Parameter:
#   AG
# Output:
#   TRUE if the node is eligible
#   TRUFALSE if the node is not eligible
##############################################################################
function can_i_be_the_master
{
    AG=$1
    my_role=$2
    start_file=$3
    ag_pkg_name=$4
    typeset -i ASYNC=$TRUE
    typeset -i final_ret=$TRUE

    typeset -i commit_mode
    typeset -i do_i_lag
    typeset -i db_health
    typeset -i max_seq_number
    typeset -i _is_prev_primary_async
    typeset -i _is_db_sync_state_healthy
    typeset -i _ret
    typeset _local_output=$mssql_tmp_dir/cmexec_remote_master_out.$$_"$AG"

    # required for do_cmexec
    typeset _output=$mssql_tmp_dir/cmexec.${0##*/}.instance_role_"$AG"
    typeset _stderr=$mssql_tmp_dir/cmexec.err.${0##*/}.instance_role_"$AG"

    # 1. check if I'm an ASYNC Replica. Give up now if so.
    get_commit_mode_from_config $LOCAL_HOST_NAME
    commit_mode=$?
    if [[ $commit_mode -eq $ASYNC ]]; then
        sg_log 0 "INFO: For Availability Group:\"$AG\" node $LOCAL_HOST_NAME, is an"\
                 "ASYNCHRONOUS_COMMIT replica, not eligible to be promoted as PRIMARY REPLICA"\
                 "Automatically."
        # Lets not exit here if i'm ASYNC and the packages on all the nodes are starting now
        # We can print a meaningful message in the package logs if we do not exit here
        if [[ ! -f $start_file ]]; then
            return $FALSE
        fi
    fi

    read_the_file_params_for_prev_primary_and_sequence_num $foot_print_file

    # 2. check if i lag from the file.
    do_i_lag_in_sec_from_file $foot_print_file $LOCAL_HOST_NAME
    do_i_lag=$?
    if [[ $do_i_lag -eq $TRUE ]]; then
        sg_log 0 "INFO: For Availability Group:\"$AG\" node $LOCAL_HOST_NAME,"\
                 "lags from the previous known PRIMARY, not eligible to be promoted as PRIMARY"\
                 "REPLICA."
        return $FALSE
    fi

    # before we go further let us get the DB values and populate in the local variables.
    get_primary_replica_identification_params $AG

    # 3. Validate the DB health parameters exit if inappropriate
    are_all_db_in_healthy_state_for_a_node $AG $LOCAL_HOST_NAME
    db_health=$?
    if [[ $db_health -eq $FALSE ]]; then
        sg_log 0 "INFO: For Availability Group:\"$AG\" node $LOCAL_HOST_NAME is not eligible"\
                 "to be promoted as PRIMARY REPLICA since one or more of the Database's"\
                 "is(are) suspended."
        check_if_last_known_primary_is_async
        _is_prev_primary_async=$?
        if [[ $_is_prev_primary_async -eq $TRUE ]]; then
             print_async_replica_is_latest_and_need_manual_failover_message $AG
        fi
        return $FALSE
    fi

    # 4. If the last known primary is ASYNCH replica, do not proceed further to failover an AG.
    check_if_last_known_primary_is_async
    _is_prev_primary_async=$?
    if [[ $_is_prev_primary_async -eq $TRUE ]]; 
    then
        print_async_replica_is_latest_and_need_manual_failover_message $AG
        return $FALSE
    fi

    # 5. Check the state of db synchronization state from the file. If the db
    # synchronization_health is not healthy, the failover command will fail. So we should not 
    # attempt to promote on this node.
    if [[ -s $foot_print_file ]]; then
        get_synchronization_health_from_file $foot_print_file $LOCAL_HOST_NAME
        _is_db_sync_state_healthy=$?
        if [[ $_is_db_sync_state_healthy -eq $FALSE ]]; then
            sg_log 0 "INFO: For Availability Group:\"$AG\" node $LOCAL_HOST_NAME is not eligible"\
                  "to be promoted as PRIMARY REPLICA since one(or more) of the Database(s)"\
                  "synchronization_health is not healthy."
            return $FALSE
        fi
    fi

    # 6. COMPARE MY SEQ NUMBER WITH THE ONE IN FILE
    get_seq_num_from_file $foot_print_file
    if [[ $seq_num_from_db -lt $SEQ_NUM_FROM_FILE ]]; then
        # I dont have updated data
        sg_log 0 "INFO: For Availability Group:\"$AG\" node $LOCAL_HOST_NAME is not eligible to"\
                 "be promoted as PRIMARY REPLICA since the sequence_number does not match with the"\
                 "previous PRIMARY REPLICA."
        return $FALSE
    fi

    # TO begin with assume i am the maximum
    max_seq_number=$seq_num_from_db
    get_nodeid_from_nodename $LOCAL_HOST_NAME
    max_node_id=$?


    # 7. Need to check if other replicas are better than me
    up_nodes=`get_pkg_up_and_detached_nodes $ag_pkg_name`
    if [[ -f $start_file ]]; then
        # If the file exists it means that package was  halted/failed and are starting now.
        # It is important to check the ASYNC replicas too in this case, there are chances that it
        # might have been PRIMARY before the package was halted/failed.
        nodes_to_be_verified="$up_nodes"
    else
        # Lets just worry about the SYNC nodes for now, there was a PRIMARY
        nodes_to_be_verified=`get_list_of_all_up_nodes_in_sync_state $up_nodes`
    fi

    for node in $nodes_to_be_verified
    do
        if [[ "$node" != "$LOCAL_HOST_NAME" ]]; then
            exec_file=$SGSBIN/mssql/hamssqlaoai_master_eligibility.sh
            cmd_opts="$node $_output $_stderr"
            cmd_args="$SG_PACKAGE_NAME $AG $max_seq_number $max_node_id $my_role"
            do_cmexec $cmd_opts $exec_file $cmd_args >$_local_output 2>&1
            if [ $? -eq 0 ]; then
                _ret=`cat "$_local_output"`
                if [[ $_ret -eq $FALSE ]] || [[ $_ret -ne $TRUE ]];then
                    final_ret=$FALSE
                    sg_log 0 "INFO: For Availability Group:\"$AG\" response from node \"$node\""\
                             "that node \"$LOCAL_HOST_NAME\" is not eligible to be promoted as"\
                             "PRIMARY REPLICA."
                    break
                fi
            else
               final_ret=$FALSE
               sg_log 0 "INFO: For Availability Group:\"$AG\" response from node \"$node\""\
                        "that node \"$LOCAL_HOST_NAME\" is not eligible to be promoted as PRIMARY"\
                        "PRIMARY REPLICA."
            fi
        fi
    done
    rm -f $_local_output
    if [[ $final_ret -eq $TRUE ]] && [[ $commit_mode -eq $ASYNC ]] && [[ -f $start_file ]]; then
        # We have to return FALSE since all the nodes are UP and ASYNC is the greatest
        # Promotion of ASYNC to PRIMARY is always Manual failover.
        print_async_replica_is_latest_and_need_manual_failover_message $AG
        return $FALSE
    else
        return $final_ret
    fi
}

function print_async_replica_is_latest_and_need_manual_failover_message
{
    AG=$1
    sg_log 0 "INFO: For Availability Group:\"$AG\", the node \"$LOCAL_HOST_NAME\" which is an"\
             "ASYNCHRONOUS_COMMIT replica was the last known PRIMARY."
    sg_log 0 "Failover to and from the ASYNCHRONOUS_COMMIT replica is always Manual."
    sg_log 0 "Refer the $DOC_NAME on how to do a Manual Failover."
    echo "For Availability Group:\"$AG\",the last known primary is an ASYNCHRONOUS replica."\
         "Failover to and from the node \"$LOCAL_HOST_NAME\" which is an ASYNCHRONOUS_COMMIT replica"\
         "is always Manual. Refer the $DOC_NAME on how to do a Manual Failover. " >> /tmp/email_body_$$.txt
    sg_send_mail "Failover to and from the ASYNCHRONOUS_COMMIT replica is always Manual." "/tmp/email_body_$$.txt"
    rm -f /tmp/email_body_$$.txt
}

##############################################################################
# Function: check_if_last_known_primary_is_async
# Output:
#   TRUE if it is ASYNC commit mode
#   FALSE if it is SYNC commit mode
# This function is used to check if the availbility mode of last known
# primary replica is ASYNCHRONOUS COMMIT
##############################################################################
function check_if_last_known_primary_is_async
{
    typeset -i commit_mode_of_prev_primary
    typeset -i ret=$FALSE

    # Retrieve the availability mode of the previous primary node.
    get_commit_mode_from_config $PRIM_NODE_FROM_FILE
    commit_mode_of_prev_primary=$?

    if [ $commit_mode_of_prev_primary -eq $TRUE ];
    then
         ret=$TRUE
    fi
    return $ret
}

function validate_required_synchronous_copies_to_commit
{
    AG=$1
    ag_pkg_name=$2
    _up_nodes=`get_pkg_up_and_detached_nodes $ag_pkg_name`
    typeset -i req_copies_to_commit=`$cmviewcl -vf line -s config -p $ag_pkg_name | \
        grep "required_synchronized_secondaries_to_commit" | awk -F"[=]" '{print $2}'`
    typeset -i sync_up_nodes_count
    

    if [ $req_copies_to_commit -eq 0 ]; then
        # REQUIRED_SYNCHRONIZED_SECONDARIES_TO_COMMIT is set to 0
        return
    fi
    sync_up_nodes_count=`get_list_of_all_up_nodes_in_sync_state $_up_nodes | wc -w`
    if [ $req_copies_to_commit -le `expr $sync_up_nodes_count - 1` ]; then
        sg_log 0 "INFO: For Availability Group:\"$AG\", the configured value of "\
                 "REQUIRED_SYNCHRONIZED_SECONDARIES_TO_COMMIT is \"$req_copies_to_commit\", the"\
                 "available SYNCHRONOUS_COMMIT replicas are \"`expr $sync_up_nodes_count - 1`\""
        sg_log 0 "INFO: High Availability and Data Protection is guaranteed."
    else
        sg_log 0 "WARNING: For Availability Group:\"$AG\", the configured value of "\
                 "REQUIRED_SYNCHRONIZED_SECONDARIES_TO_COMMIT is \"$req_copies_to_commit\", the"\
                 "available SYNCHRONOUS_COMMIT replicas are \"`expr $sync_up_nodes_count - 1`\""
        sg_log 0 "WARNING: Applications cannot connect to the database till the"\
                 "REQUIRED_SYNCHRONIZED_SECONDARIES_TO_COMMIT is satisfied."
        echo "For Availability Group:\"$AG\", the configured value of "\
                 "REQUIRED_SYNCHRONIZED_SECONDARIES_TO_COMMIT is \"$req_copies_to_commit\", the"\
                 "available SYNCHRONOUS_COMMIT replicas are \"`expr $sync_up_nodes_count - 1`\""\
                 "Applications cannot connect to the database till the"\
                 "REQUIRED_SYNCHRONIZED_SECONDARIES_TO_COMMIT is satisfied." >> /tmp/email_body_$$.txt
        sg_send_mail "Configured value of REQUIRED_SYNCHRONIZED_SECONDARIES_TO_COMMIT is not satisfied for AG \"$AG\"" "/tmp/email_body_$$.txt"
        rm -f /tmp/email_body_$$.txt
    fi
}

#This function is used to send the notification if there is no
#primary node eligible for failover of AG
function send_notification_if_no_primary
{
    typeset ag_pkg_name="$1"
    typeset up_nodes=`get_pkg_up_and_detached_nodes $ag_pkg_name`
    typeset AG="$2"
    typeset -i _node_id
    typeset _node_id_array
    typeset -i _lowest_node_id

    #Get the nodeid from all nodes to retrieve the node with lowest node id.
    for node in $up_nodes
    do
       get_nodeid_from_nodename $node
       _node_id=$?
       _node_id_array+=($_node_id)
    done
    _lowest_node_id=$(printf "%s\n" "${_node_id_array[@]}" | sort -u | head -1)
    get_node_name_from_nodeid $_lowest_node_id
    _node_with_lowest_nodeid=`get_node_name_from_nodeid $_lowest_node_id`

    if [[ "$_node_with_lowest_nodeid" == "$LOCAL_HOST_NAME" ]];
    then
        echo "The failover of availability group $AG is unsuccessful"\
             "as none of the seondary replica instances is eligible to be primary." >> /tmp/email_body_$$.txt
        sg_send_mail "The failover of availability group $AG is unsuccessful." "/tmp/email_body_$$.txt"
        rm -f /tmp/email_body_$$.txt
    fi
}

##############################################################################
# Function: retrieve_and_set_gr_local_replica_role
# Input Parameter:
#   AG
#   Name of the GR
# Output:
#   NONE
##############################################################################
function retrieve_and_set_gr_local_replica_role
{
    AG=$1
    ROLE_GR=$2
    typeset -i mssql_ag_role
    typeset -i sg_ag_role

    get_the_role_of_the_ag_on_local_replica $AG
    mssql_ag_role=$?
    map_sql_server_ag_role_to_sg_ag_role $mssql_ag_role
    sg_ag_role=$?
    set_gr_role_on_local_node "$ROLE_GR" $sg_ag_role $AG

    return $mssql_ag_role
}

##############################################################################
# Function: promote_from_resolving_to_secondary
# Input Parameter:
#   AG
# Output:
#   NONE
##############################################################################
function promote_from_resolving_to_secondary
{
    AG=$1
    # I'm RESOLVING, because I failed when I was PRIMARY mostly...
    sg_log 0 "INFO: The Availability Group:\"$AG\" on node $LOCAL_HOST_NAME is in RESOLVING state"
    sg_log 0 "INFO: Since there is already a PRIMARY REPLICA in the cluster node $LOCAL_HOST_NAME"\
             "will be attempted to be promoted as SECONDARY"
    set_role_to_secondary $AG
    sg_log 0 "SUCCESS: The Availability Group:\"$AG\" on node $LOCAL_HOST_NAME is promoted as"\
             "SECONDARY REPLICA."
}

function get_package_state_from_package_name
{
    echo `$cmviewcl -f line | awk -F= '/^package:'$1'\|state=/ {print $2}'`
}

function get_write_intent_package_name
{
    # From the dependency we can get the package name
    # After awk we get this
    #    package:PKG1|dependency:PKG1_DEP|condition=APKG = up
    # After cut we get this
    #    package:PKG1
    # After sed we get this
    #    PKG1

    _pkg_name=$1

    echo `$cmviewcl -vf line -s config |\
          awk '/^package:[^|]*\|dependency:[^|]*\|condition='$_pkg_name'/' |\
          cut -d "|" -f 1 | sed 's/package://'`
}

function check_if_there_is_primary_ag_from_the_db
{
    ag_name=$1
    ag_pkg_name=$2
    up_nodes=`get_pkg_up_and_detached_nodes $ag_pkg_name`
    typeset -i found=0

    # required for do_cmexec
    typeset _output=$mssql_tmp_dir/cmexec.${0##*/}.instance_role_"$AG"
    typeset _stderr=$mssql_tmp_dir/cmexec.err.${0##*/}.instance_role_"$AG"

    cmd="$SQLCMD"
    args="-U $MSSQL_MSSQL_ADMIN -P $MSSQL_MSSQL_PASSWORD -Q \"set nocount on
          SELECT
              dhars.role,
              dhars.role_desc
          FROM
              master.sys.dm_hadr_availability_replica_states dhars, sys.availability_groups ag
          WHERE
              ag.name = '$ag_name' AND
              dhars.is_local = 1 AND
              dhars.group_id = ag.group_id;\" -W -h -1"
    for node in $up_nodes
    do
        cmd_opts="$node $_output $_stderr"
        echo `do_cmexec $cmd_opts $cmd $args` | grep -qw PRIMARY
        db_ret=$?
        if [ $db_ret -eq 0 ]; then
            found=1
            break
        fi
    done
    if [ $found -eq 1 ]; then
        return $TRUE
    else
        return $FALSE
    fi
}

##############################################################################
# Function: check_if_there_is_primary_for_ag
# To check if there is a primary for a given AG we can check if
# the write intent package configured for the given AG is running.
# If there in no write intent package check the DB
# Output:
#        0 - TRUE
#        1 - FALSE
##############################################################################
function check_if_there_is_primary_for_ag
{
    AG=$1
    AG_PKG_NAME=$2

    write_intent_pkg=`get_write_intent_package_name $AG_PKG_NAME`

    _state=`get_package_state_from_package_name $write_intent_pkg`
    if [[ "$_state" = "running" ]]; then
        return $TRUE
    else
        # Should we check the DB if there is a PRIMARY.
        # The write intent package might be down due to user halt.
        check_if_there_is_primary_ag_from_the_db $AG $AG_PKG_NAME
        ret=$?
        return $ret
    fi
}

##############################################################################
# Function: get_primary_replica_identification_params
# Uses the sqlcmd to on each replica node and get the following
#    Column Value                   Table
# 1. sequence_number       : sys.availability_groups
# 2. role                  : master.sys.dm_hadr_availability_replica_states
# 3. availability_mode     : master.sys.availability_replicas
# 4. recovery_health       : master.sys.dm_hadr_availability_replica_states
# 5. secondary_lag_seconds : sys.dm_hadr_database_replica_states
# Input: AG name
# Output: Above listed values.
##############################################################################
function get_primary_replica_identification_params
{
    ag_name=$1
    typeset capture_once=0
    unset seq_num_from_db
    unset role_from_db
    unset availability_mode_from_db
    unset recover_health_from_db
    unset db_is_suspended_from_db
    unset db_names_from_db

    cmd="$SQLCMD -U $MSSQL_MSSQL_ADMIN -P $MSSQL_MSSQL_PASSWORD -Q \"set nocount on SELECT
            ag.$seq_string,
            ars.role,
            ar.availability_mode,
            ars.recovery_health,
            drs.is_suspended,
            adc.database_name
        FROM sys.dm_hadr_database_replica_states drs, sys.availability_groups ag,
        master.sys.availability_replicas ar, master.sys.dm_hadr_availability_replica_states ars,
        master.sys.availability_databases_cluster adc
        WHERE
            ag.name = '$ag_name' AND
            ars.is_local = 1 AND
            drs.group_id = ag.group_id AND
            drs.group_database_id = adc.group_database_id AND
            ar.replica_id = drs.replica_id AND
            ar.replica_id = ars.replica_id;\" -W"

    output=`execute_sql_cmd_on_local_node "$seq_string" $cmd`
    if [[ $output != "CMD_FAILED" ]]; then
        while read -r line
        do
            if echo "$line" | grep -q -e "^$seq_string" -e "^----------"
            then
                continue
            else
                if [ $capture_once -eq 0 ]; then
                    seq_num_from_db=`echo "$line" | awk '{print $1}'`
                    role_from_db=`echo "$line" | awk '{print $2}'`
                    availability_mode_from_db=`echo "$line" | awk '{print $3}'`
                    capture_once=1
                fi
                recover_health_from_db+=(`echo "$line" | awk '{print $4}'`)
                db_is_suspended_from_db+=(`echo "$line" | awk '{print $5}'`)
                db_names_from_db+=(`echo "$line" | awk '{print $6}'`)
            fi
        done <<<"$output"
    fi
}

function get_ag_pkg_name_from_ag_name
{
    AG=$1
    echo `$cmviewcl -vf line -s config | \
    awk '/^package:[^|]*\|sgemssql\/mssqlaoai\/availability_group:'$AG'\|sgemssql\/mssqlaoai\/availability_group='$AG'/' | \
    cut -d "|" -f 1 | sed 's/package://'`
}

function get_package_up_nodes
{
    _pkg_name=$1
    echo `$cmviewcl -f line -p $_pkg_name | awk '/^node:[^|]*\|status=up/' |\
              cut -d "|" -f 1 | sed 's/node://'`
}

function get_package_detached_nodes
{
   _pkg_name=$1
    echo `$cmviewcl -f line -p $_pkg_name | awk '/^node:[^|]*\|status=detached/' |\
          cut -d "|" -f 1 | sed 's/node://'`

}

# This function gets both upnodes and detached nodes and combines the list
# without duplicates
function get_pkg_up_and_detached_nodes
{
    _pkg_name=$1
    up_nodes=`get_package_up_nodes $_pkg_name`
    detached_nodes=`get_package_detached_nodes $_pkg_name`
    if [[ -n "${detached_nodes[@]}" ]]; then
        eligible_nodes=( "${up_nodes[@]}" "${detached_nodes[@]}" )
    else
        eligible_nodes=( "${up_nodes[@]}" )
    fi
    e_nodes=`echo "${eligible_nodes[*]}" | tr ' ' '\n' | sort -u | tr '\n' ' '`
    echo "${e_nodes[*]}"
}


function get_nodes_configured_for_the_package
{
    _pkg_name=$1
    echo `echo "$VCLVIEW_SCONFIG" | awk -F= '/^package:'$_pkg_name'\|node:[^|]*\|name=/ {print $2}'`
}

function check_if_all_the_configured_nodes_are_available
{
    _pkg_name=$1
    _up_nodes=`get_package_up_nodes $_pkg_name`
    typeset -i _up_node_count
    typeset -i _configured_pkg_nodes_count

    _up_node_count=`echo "$_up_nodes" | wc -w`
    _configured_pkg_nodes_count=`get_nodes_configured_for_the_package $_pkg_name | wc -w`

    if [ $_up_node_count -eq $_configured_pkg_nodes_count ]; then
        return $TRUE
    fi
    return $FALSE
}

function print_no_quorum_message
{
    _pkg_name=$1
    AG=$2
    sg_log 0 "INFO: All the nodes configured in the package \"$_pkg_name\" for the Availability"\
             "Group \"$AG\" are not up"
    sg_log 0 "INFO: For Automatic failover all the nodes configured for the Availability Group"\
             "should be up, to identify the latest and greatest replica among the configured ones."
    sg_log 0 "To rectify this issue, you can choose to ..."
    sg_log 0 "1. Start the package \"$_pkg_name\" on all the nodes using cmrunpkg"
    sg_log 0 "2. Refer the $DOC_NAME on how to do a Manual Failover."
}

###############################################################################
# This function initializes all the global variables used to hold the cached
# data. This function should be called in the beginning of the main scripts
# that use cached data
# Internal Use only
#
# Arguments:  $1 : Site Controller Package Name
# Returns:    none
###############################################################################
function init_cache
{
    set_retry_interval

    # Cache cmviewcl -v fline -s config output
    cache_clview_sconfig

    # Cache cmviewcl -f line output
    cache_clview
}

##########################################################################
# This function is used to set retry interval.
##########################################################################
function set_retry_interval
{
    clview_retry_interval=50 #twice the recommended average cluster member timeout
    clview_retry_times=3
    cmexec_retry_interval=15
}

###############################################################################
# This function executes a command on the specified node using cmexec command
# and stores its output in the variable CMEXEC.
#
# Internal Use only
#
#
# Arguments:  $1   : node name
#             $2   : command
#             $3.. : arguments to the command
#
#
# Returns:    none
#
###############################################################################
function do_cmexec
{
    typeset _node_name=$1
    typeset _output=$2
    typeset _stderr=$3
    typeset _cmd=$4
    typeset _args=
    typeset _out
    typeset -i _rc=0
    typeset _line
    typeset _node_status
    typeset -i counter=$clview_retry_times
    # The variables _output _stderr are initialized in the monitoring script, to avoid the
    # name space collision

    shift 4
    while [ $# -gt 0 ]; do
        if [ -z "$1" ]; then
            _args="$_args \"\""
        else
            _args="$_args $1"
        fi
        shift
    done

    $cmexec $_node_name -o $_output -t $cmexec_retry_interval $_cmd $_args >/dev/null 2>$_stderr
    grep -q ^cmexec: $_stderr
    if [ $? -eq 0 ]; then
        while [[ $counter -gt 0 ]]
        do
            # Refresh cmviewcl cache
            cache_clview
            _node_status=`echo "$VCLVIEW" | awk -F "[:|=]" '(($1 == "node") && \
                        ($2 == "'$_node_name'") && ($3 == "status")){print $4}'`

            if [[ "$_node_status" == "up" ]]; then
                _out="$(cat $_output)"
                #sleep for retry interval
                sleep $clview_retry_interval
                $cmexec $_node_name -o $_output -t $cmexec_retry_interval $_cmd $_args >/dev/null 2>$_stderr
                grep -q ^cmexec: $_stderr
                if [[ $? -ne 0 ]]; then #if cmexec not failed
                    break
                fi
                counter=`expr $counter - 1`
            fi
        done

        if [[ $counter -eq 0 ]]; then
            _out="$(cat $_output)"
            sg_log 0 "_out"
            rm -f $_output 2>/dev/null
            rm -f $_stderr 2>/dev/null
            return $_cmexec_err
        fi
    fi

    _out="$(cat $_output)"
    echo "$_out" | \
    {
    read _rc
    while read -r _line; do
        if [ $_rc -eq 0 ]; then
            echo "$_line"
        else
            echo "$_line" >&2
        fi
    done
    rm -f $_output 2>/dev/null
    rm -f $_stderr 2>/dev/null
    return $_rc
    }
}


# This function will be called by hamssqldb.sh during online modification
# to verify the attributes that are modified are supported for online
# modifications or not.

function verify_unsupported_online_modification
{
    typeset -i flag=0
    config_type=$1  #this can be either aofi or aoai

    if [[ "$config_type" = "aofi" || "$config_type" = "aoai" ]]; then

        if [[ "${MSSQL_MSSQL_PORT}" != "${PREV_MSSQL_MSSQL_PORT}" ]]; then
            sg_log 0 "ERROR: Online modification of MSSQL_PORT attribute value is not supported"
            flag=1
        fi

        if [[ $flag -eq 1 ]]; then
            exit 1
        fi
    fi

}

